import sys
import signal
from copy import deepcopy
from collections import Counter
from multiprocessing import Queue

from cml.shared.logger import ConstructionInfo, ConstLogger
from cml.domain.construction_parallel import BackgroundConstructor


__all__ = (
    "Constructor",
)


def handle_background_constructor_failure(signum, stack):
    sys.stdout.write("An exception occured in the background constructor. "
                     "Tried to label a learnblock. "
                     "Refer to stderr to identify the exception cause.")
    sys.exit()
signal.signal(signal.SIGALRM, handle_background_constructor_failure)


class Constructor:
    def __init__(self, mode, ml_models, settings):
        self.mode = mode
        self.settings = settings
        self.ml_models = ml_models
        self._construction = self._init_construction(mode)
        self.logger = None
        self.__process = None
        self.__input_queue = None
        self.__output_queue = None

    @ConstLogger
    def construct(self, learnblock):
        for log_info, block in self._construction(learnblock):
            yield log_info, block

    @ConstLogger
    def construct_parallel(self, learnblock):
        self.prepare_background_process()
        self.__input_queue.put(learnblock)
        yield self.__output_queue.get()
        while not self.__output_queue.empty():
            yield self.__output_queue.get()

    @property
    def max_target_error(self):
        return self.settings.max_target_error

    @max_target_error.setter
    def max_target_error(self, value):
        self.settings.max_target_error = value

    @property
    def max_model_targets(self):
        return self.settings.max_model_targets

    @max_model_targets.setter
    def max_model_targets(self, value):
        self.settings.max_model_targets = value

    @property
    def min_category_size(self):
        return self.settings.min_category_size

    @min_category_size.setter
    def min_category_size(self, value):
        self.settings.min_category_size = value
        self._init_construction(self.mode)

    @property
    def max_categories(self):
        return self.settings.max_categories

    @max_categories.setter
    def max_categories(self, value):
        self.settings.max_categories = value
        self._init_construction(self.mode)

    def prepare_background_process(self):
        if not self.__process or not self.__process.is_alive():
            self.__input_queue = Queue()
            self.__output_queue = Queue()
            background = BackgroundConstructor(self.ml_models,
                                               self.logger,
                                               self.settings,
                                               self.__input_queue,
                                               self.__output_queue)
            background.name = "BackgroundConstructor"
            background.daemon = True
            background.start()
            self.__process = background

    def _init_construction(self, mode):
        if mode == "conceptual":
            return self._construct_conceptual_knowledge
        else:
            return self._construct_procedural_knowledge

    def _construct_conceptual_knowledge(self, learnblock):
        logger_info = ConstructionInfo()

        if learnblock.labeled():
            logger_info.state = ConstructionInfo.State.PASSED
            yield (logger_info, learnblock)

        else:
            for ml_model in self.ml_models:
                for cluster_number in range(2, self.settings.max_categories+1):
                    ml_model.cluster = cluster_number
                    trained_model = ml_model.train(learnblock)

                    found_clusters = len(Counter(trained_model.get_labels()).keys())
                    if found_clusters != cluster_number:
                        break

                    for cluster, size in trained_model.cluster_sizes.items():
                        if size < self.settings.min_category_size:
                            break

                    else:
                        labeled = self.label_block(learnblock, trained_model)

                        # Fill the logger object
                        logger_info.n_cluster = cluster_number
                        logger_info.subject = ml_model.subject
                        logger_info.state = ConstructionInfo.State.CONSTRUCTED

                        yield (logger_info, labeled)

            logger_info.state = ConstructionInfo.State.DISCARDED
            yield (logger_info, None)

    def label_block(self, block, trained_model):
        copied_block = deepcopy(block)
        labels = trained_model.get_labels()
        copied_block.set_column("Z", labels)
        copied_block.n_cluster = trained_model.cluster
        copied_block.origin = block.origin
        subject = "{}{:02}".format(trained_model.abbreviation, trained_model.cluster)
        copied_block.subject = subject
        return copied_block

    def _construct_procedural_knowledge(self,
                                        learnblock,
                                        procedural_complexity=None,
                                        max_target_error=None):
        if learnblock.labeled:
            yield learnblock
        else:
            for ml_model in self.ml_models:
                for target_number in range(2, procedural_complexity+1):
                    model = ml_model()
                    model.io_shape = learnblock.learn_cols
                    model.target_number = target_number
                    trained_model = model.train(learnblock.as_numpy_array())
                    if trained_model.target_error < max_target_error:
                        for labels in trained_model.targets:
                            labeled_learnblock = learnblock.set_labels(list(labels))
                            labeled_learnblock.n_cluster = target_number
                            yield labeled_learnblock


#     learnblock, model, min_category_size = args[0]
#     if learnblock.labeled:
#         return learnblock
#
#     else:
#         trained_model = model.train(learnblock)
#         for cluster, size in trained_model.cluster_sizes.items():
#             if size < min_category_size:
#                 return
#         else:
#             labels = trained_model.get_labels()
#             labeled_learnblock = learnblock.set_labels(labels)
#             labeled_learnblock.n_cluster = model.cluster
#             purpose = "{}{:02}".format(model.abbreviation, model.cluster)
#             labeled_learnblock.purpose = purpose
#             return labeled_learnblock
