from os import mkdir
from os.path import expanduser, join
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler
from collections import namedtuple

from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain.knowledge import TS_QUEUE
from cml.shared.errors import (
    NoModelReconstructedError,
    NotReducableLearnblockWarning,
    ModiError)


Components = namedtuple(
    "Components",
    "source constructor selector reconstructor deconstructor")


class KnowledgeSearchUsecase(Usecase):
    FIRST_TIER = 1

    def __init__(self):
        super().__init__()

    def process(self, request):
        if not self.valid_modi(request): raise ModiError()
        components = self.init_loggers(request)
        self.init_components(components)
        self.run(components, request.parallel)
        return components.deconstructor.knowledge_database

    def run(self, components, parallel):
        source, constructor, selector, reconstructor, deconstructor = components

        if parallel:
            construct = constructor.construct_parallel
            reconstruct = reconstructor.reconstruct_parallel
        else:
            construct = constructor.construct
            reconstruct = reconstructor.reconstruct

        tier = self.FIRST_TIER

        deconstructor.logger.protocol("{:#^100}".format("TIER 1"))
        for learnblock in source.learnblocks:
            for complete_learnblock in construct(learnblock):
                try:
                    reduced = selector.select(complete_learnblock)
                    pragmatic = reconstruct(tier, reduced)
                    deconstructor.deconstruct(tier, pragmatic, reduced)

                except NotReducableLearnblockWarning as error:
                    continue

                except NoModelReconstructedError as error:
                    continue

        self.run_on_high_tiers(construct, reconstruct, deconstructor)

    def run_on_high_tiers(self, construct, reconstruct, deconstructor):
        tier = self.FIRST_TIER

        while 1:
            try:
                new_tier, learnblock = TS_QUEUE.pop()
                if new_tier > tier:
                    banner = "TIER {}".format(new_tier)
                    deconstructor.logger.protocol("{:#^100}".format(banner))
                    tier = new_tier

                for complete_learnblock in construct(learnblock):
                    try:
                        pragmatic = reconstruct(tier, complete_learnblock)
                        deconstructor.deconstruct(tier, pragmatic,
                                                  complete_learnblock)

                    except NotReducableLearnblockWarning as error:
                        continue

                    except NoModelReconstructedError as error:
                        continue

            except IndexError:  # deque is empty
                break

    def valid_modi(self, request):
        return request.constructor.mode == request.reconstructor.mode == \
               request.deconstructor.mode

    def init_loggers(self, request):
        source = request.deconstructor.source
        log_dir = create_log_dir()
        iteration_log_file = join(log_dir, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")
        iteration_logger.addHandler(iteration_handler)

        source.logger = iteration_logger
        request.constructor.logger = iteration_logger
        request.feature_selector.logger = iteration_logger
        request.reconstructor.logger = iteration_logger
        request.deconstructor.logger = iteration_logger
        request.deconstructor.knowledge_database.inject_logger(iteration_logger)

        if request.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)

        components = Components(source=source,
                                constructor=request.constructor,
                                selector=request.feature_selector,
                                reconstructor=request.reconstructor,
                                deconstructor=request.deconstructor)
        return components

    def init_components(self, components):
        components.reconstructor.next_model_counter = \
            components.deconstructor.next_model_counter
        components.deconstructor.reconstructor = components.reconstructor


def create_log_dir():
    log_dir = join(expanduser('~'),
                   ".cml",
                   datetime.now().strftime("%d%m%Y%H%M%S"))
    mkdir(log_dir)
    return log_dir

